/*
 * Copyright 2009 Ben Axelrod
 * 
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the Creative Commons Attribution-Noncommercial-Share
 * Alike 3.0 United States License. 
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * 
 * To view a copy of this license, visit:
 * 
 * http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 *
 * or send a letter to:
 * Creative Commons
 * 171 Second Street, Suite 300
 * San Francisco, California 94105 USA
 */


using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Xml;

using Microsoft.Ccr.Core;
using Microsoft.Dss.Core;
using Microsoft.Dss.Core.Attributes;
using Microsoft.Dss.ServiceModel.Dssp;
using Microsoft.Dss.ServiceModel.DsspServiceBase;
using Microsoft.Robotics.PhysicalModel;

// for Http Operations
using System.Net;
using System.Collections.Specialized;
using Microsoft.Dss.Core.DsspHttp;
using Microsoft.Dss.Core.DsspHttpUtilities;
using System.IO;
using System.Drawing;
using System.Drawing.Imaging;
using W3C.Soap;
using System.Net.Mime;

using sicklrf = Microsoft.Robotics.Services.Sensors.SickLRF.Proxy;
using pos2d = Microsoft.Robotics.Services.SimplePosition.Proxy;

namespace Robotics.Mapper
{

    /// <summary>
    /// The Mapper Service
    /// </summary>
    [DisplayName("Mapper")]
    [Description("The Mapper Service")]
    [Contract(Contract.Identifier)]
    public class MapperService : DsspServiceBase
    {
        /// <summary>
        /// The state value for the Mapper service
        /// </summary>
        private MapperState _state = new MapperState();

        /// <summary>
        /// The main port for the Mapper service
        /// </summary>
        [ServicePort("/mapper", AllowMultipleInstances = false)]
        private MapperOperations _mainPort = new MapperOperations();

        [EmbeddedResource("Robotics.Mapper.Mapper.xslt")]
        string _transform = null;

        DsspHttpUtilitiesPort _httpUtilities = new DsspHttpUtilitiesPort();

        [Partner("/simpleposition", Contract = pos2d.Contract.Identifier, CreationPolicy = PartnerCreationPolicy.UseExisting)]
        pos2d.SimplePositionOperations _positionPort = new pos2d.SimplePositionOperations();

        [Partner("sickLaser", Contract = sicklrf.Contract.Identifier, CreationPolicy = PartnerCreationPolicy.UseExisting)]
        sicklrf.SickLRFOperations _sickPort = new sicklrf.SickLRFOperations();
        sicklrf.SickLRFOperations _sickNotify = new sicklrf.SickLRFOperations();

        private const float xOffsetmm = 1524; //mm
        private const float zOffsetmm = 762; //mm
        private const float xOffsetm = 1.524f; //m
        private const float zOffsetm = 0.762f; //m

        private const int imgWidth = 710; //pix
        private const int imgHeight = 600; //pix

        private const float scaleppm = 50; //pixels per meter
        private const float scaleppmm = 0.05f; //pixels per mm

        private const int robothalfwidth = 10; //pixels

        private const int savePointsEveryNscans = 10;

        private const int maxLaserDist = 8000;

        /// <summary>
        /// Default Service Constructor
        /// </summary>
        public MapperService(DsspServiceCreationPort creationPort) : base(creationPort)
        {
        }

        /// <summary>
        /// Service Start
        /// </summary>
        protected override void Start()
        {
            _httpUtilities = DsspHttpUtilitiesService.Create(Environment);

            base.Start();

            _state.AllDistanceMeasurements = new List<PointList>();

            // Set up an interleave for notifications from sick laser
            Activate(new Interleave(
                new ExclusiveReceiverGroup(
                    Arbiter.Receive<sicklrf.Replace>(true, _sickNotify, SickNotifyReplaceHandler)
                ),
                new ConcurrentReceiverGroup())
            );

            //subscribe to sick laser
            Arbiter.Choice(
                _sickPort.Subscribe(_sickNotify),
                delegate(SubscribeResponseType response) { },
                    delegate(Fault fault)
                    {
                        LogError(null, "Unable to subscribe to Sick", fault);
                    }
            );
        }

        /// <summary>
        /// Handle get requests that come in on the dssp.tcp port
        /// </summary>
        /// <param name="get"></param>
        /// <returns></returns>
        [ServiceHandler(ServiceHandlerBehavior.Concurrent)]
        public virtual IEnumerator<ITask> GetHandler(Get get)
        {
            get.ResponsePort.Post(_state);
            yield break;
        }


        long nScans = 0;
        bool firstScan = true;

        /// <summary>
        /// Got a notification from the laser, now get the current position and add it to the state
        /// </summary>
        /// <param name="replace"></param>
        private void SickNotifyReplaceHandler(sicklrf.Replace replace)
        {
            Activate(Arbiter.Receive<pos2d.SimplePositionState>(false,
                _positionPort.Get(),
                    delegate(pos2d.SimplePositionState response)
                    {
                        float moved = (float)(_state.LatestX - response.X) +
                                      (float)(_state.LatestZ + response.Y) +
                                      (float)(_state.LatestAngle - response.Theta);

                        _state.LatestX = (float)response.X;
                        _state.LatestZ = -(float)response.Y;
                        _state.LatestAngle = (float)response.Theta;

                        _state.LatestDistanceMeasurements = new int[361];
                        if (replace.Body.DistanceMeasurements != null)
                        {
                            _state.LatestDistanceMeasurements = replace.Body.DistanceMeasurements;

                            //small optimization. don't save if we havent moved since last scan.
                            if (moved != 0 || firstScan)
                            {
                                //only save ever N scans
                                if (nScans % savePointsEveryNscans == 0)
                                {
                                    PointList newpoints = new PointList(TranslatePoints(_state.LatestX,
                                                                                        _state.LatestZ,
                                                                                        _state.LatestAngle,
                                                                                        replace.Body.DistanceMeasurements));
                                    _state.AllDistanceMeasurements.Add(newpoints);
                                }
                                nScans++;
                                firstScan = false;
                            }
                        }
                    }
            ));
        }


        private const string _root = "/mapper";
        private const string _latest = "/latest";
        private const string _all = "/all";
        private const string _raw = "/raw";

        /// <summary>
        /// Handle get requests that come in on the http port
        /// </summary>
        [ServiceHandler(ServiceHandlerBehavior.Exclusive)]
        public virtual IEnumerator<ITask> HttpGetHandler(HttpGet httpGet)
        {
            HttpListenerRequest request = httpGet.Body.Context.Request;
            HttpListenerResponse response = httpGet.Body.Context.Response;

            Stream image = null;

            string path = request.Url.AbsolutePath;

            if (path.EndsWith(_latest) || path.EndsWith(_root))
                image = GenerateLatest();
 
            else if (path.EndsWith(_all))
                image = GenerateAll();
          
            else if (path.EndsWith(_raw))
            {
                //no transform
                HttpResponseType rsp = new HttpResponseType(HttpStatusCode.OK, _state);
                httpGet.ResponsePort.Post(rsp);
            }

            if (image != null)
            {
                SendJpeg(httpGet.Body.Context, image);
            }
            else
            {
                httpGet.ResponsePort.Post(Fault.FromCodeSubcodeReason(
                    W3C.Soap.FaultCodes.Receiver,
                    DsspFaultCodes.OperationFailed,
                    "Unable to generate Image"));
            }
            yield break;
        }


        /// <summary>
        /// translates a laser scan into a List of world Points in mm
        /// </summary>
        /// <param name="robotX">the robot's world x position</param>
        /// <param name="robotZ">the robot's world z position</param>
        /// <param name="robotAngle">the robot's world angle<remarks>0 points in negative z axis</remarks></param>
        /// <param name="measurements">the int[] of mm ranges<remarks>scan right to left</remarks></param>
        /// <returns>a List&lt;Point&gt; of world coordinates in mm</returns>
        private List<Point> TranslatePoints(float robotX, float robotZ, float robotAngle, int[] measurements)
        {
            if (measurements != null)
            {
                List<Point> points = new List<Point>(360);
                double a = robotAngle;
                for (int i = 1; i <= 360; i++)
                {
                    if (measurements[i] > 0 && measurements[i] < maxLaserDist)
                    {
                        //world coords of points in mm
                        float xcoord = 1000.0f * robotX + measurements[i] * (float)Math.Cos(a);
                        float zcoord = 1000.0f * robotZ - measurements[i] * (float)Math.Sin(a);
                        points.Add(new Point((int)Math.Round(xcoord), (int)Math.Round(zcoord)));
                    }
                    a += 0.5 * Math.PI / 180.0;
                }
                return points;
            }
            return null;
        }

        /// <summary>
        /// Draw an image using only the latest laser scan
        /// </summary>
        /// <returns></returns>
        private Stream GenerateLatest()
        {

            MemoryStream memory = null;

            using (Bitmap bmp = new Bitmap(imgWidth, imgHeight))
            {
                using (Graphics g = Graphics.FromImage(bmp))
                {
                    g.Clear(Color.LightGray);

                    float dx = scaleppm * (_state.LatestX + xOffsetm);
                    float dz = scaleppm * (_state.LatestZ + zOffsetm);

                    //draw robot
                    Point robotupperleft = new Point((int)Math.Round(dx - robothalfwidth), (int)Math.Round(dz - robothalfwidth));
                    Point robotcenter = new Point((int)Math.Round(dx), (int)Math.Round(dz));
                    Point robotfront = new Point((int)Math.Round(dx - robothalfwidth * Math.Sin(_state.LatestAngle)),
                                                 (int)Math.Round(dz - robothalfwidth * Math.Cos(_state.LatestAngle)));
                    g.DrawEllipse(new Pen(Color.Red), new Rectangle(robotupperleft, new Size(2 * robothalfwidth, 2 * robothalfwidth)));
                    g.DrawLine(new Pen(Color.Red), robotcenter, robotfront);
                }
                
                //translate the latest laser scan to world coordinates
                List<Point> newpoints = new List<Point>(TranslatePoints(_state.LatestX,
                                                                        _state.LatestZ,
                                                                        _state.LatestAngle,
                                                                        _state.LatestDistanceMeasurements));

                //translate mm world coordinates to image coordinates and draw

                foreach (Point p in newpoints)
                {
                    int x = (int)Math.Round(scaleppmm * (p.X + xOffsetmm));
                    int y = (int)Math.Round(scaleppmm * (p.Y + zOffsetmm));

                    if (x > 0 && x < imgWidth && y > 0 && y < imgHeight)
                        bmp.SetPixel(x, y, Color.Black);
                }


                memory = new MemoryStream();
                bmp.Save(memory, ImageFormat.Png);
                memory.Position = 0;

            }
            return memory;
        }

        /// <summary>
        /// draw an image using all recorded laser scans
        /// </summary>
        /// <returns></returns>
        private Stream GenerateAll()
        {
            MemoryStream memory = null;

            int Nmeasurements = _state.AllDistanceMeasurements.Count;

            using (Bitmap bmp = new Bitmap(imgWidth, imgHeight))
            {
                using (Graphics g = Graphics.FromImage(bmp))
                {
                    g.Clear(Color.LightGray);

                    float dx = scaleppm * (_state.LatestX + xOffsetm);
                    float dz = scaleppm * (_state.LatestZ + zOffsetm);

                    //draw robot
                    Point robotupperleft = new Point((int)Math.Round(dx - robothalfwidth), (int)Math.Round(dz - robothalfwidth));
                    Point robotcenter = new Point((int)Math.Round(dx), (int)Math.Round(dz));
                    Point robotfront = new Point((int)Math.Round(dx - robothalfwidth * Math.Sin(_state.LatestAngle)),
                                                 (int)Math.Round(dz - robothalfwidth * Math.Cos(_state.LatestAngle)));
                    g.DrawEllipse(new Pen(Color.Red), new Rectangle(robotupperleft, new Size(2 * robothalfwidth, 2 * robothalfwidth)));
                    g.DrawLine(new Pen(Color.Red), robotcenter, robotfront);
                }

                //translate mm world coordinates to image coordinates and draw
                for( int i=0; i< Nmeasurements; i++) //each (PointList L in _state.AllDistanceMeasurements)
                {
                    foreach (Point p in _state.AllDistanceMeasurements[i].Points)
                    {
                        int x = (int)Math.Round(scaleppmm * (p.X + xOffsetmm));
                        int y = (int)Math.Round(scaleppmm * (p.Y + zOffsetmm));

                        if (x > 0 && x < imgWidth && y > 0 && y < imgHeight)
                            bmp.SetPixel(x, y, Color.Black);
                    }
                }

                memory = new MemoryStream();
                bmp.Save(memory, ImageFormat.Png);
                memory.Position = 0;

            }
            return memory;
        }

        

        private void SendJpeg(HttpListenerContext context, Stream stream)
        {
            WriteResponseFromStream write = new WriteResponseFromStream(context, stream, MediaTypeNames.Image.Jpeg);
            
            _httpUtilities.Post(write);

            Activate(
                Arbiter.Choice(
                    write.ResultPort,
                    delegate(Stream res)
                    {
                        stream.Close();
                    },
                    delegate(Exception e)
                    {
                        stream.Close();
                        LogError(e);
                    }
                )
            );
        }


    }
}
